#!/usr/bin/env python
# encoding: utf8

# Copyright 2024 D E Haynes

# This file is part of balladeer.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple
import enum
import operator

from balladeer import Drama
from balladeer.lite.loader import Loader


class Resident(Drama):

    Move = namedtuple("Move", ["heading", "spot", "via"])

    def __init__(self, *args, selector: dict[str, list] = {}, **kwargs):
        self.selector = selector | {"states": set(selector.get("states", []))}
        self.selector["paths"] = self.selector.get("paths", [])
        super().__init__(*args, **kwargs)

    @property
    def focus(self):
        selected = [i for i in self.world.typewise.get("Focus", []) if self.is_resident(i.get_state("Spot"))]
        return next(reversed(sorted(selected, key=operator.attrgetter("state"))), None)

    @property
    def exits(self):
        spot = self.focus.get_state("Spot")
        return [self.Move(*option) for option in sorted(self.world.map.options(spot), key=str)]

    def is_resident(self, *args: tuple[enum.Enum]):
        states = self.selector["states"]
        return all(str(i).lower() in states for i in args if i or states)

    def scripts(self, assets: list):
        return [
            i for i in assets
            if isinstance(i, Loader.Scene)
            and any(i.path.match(p) for p in self.selector["paths"])
            or not self.selector["paths"]
        ]



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2024 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#!/usr/bin/env python
# encoding: utf8

import pathlib
import unittest
from unittest.mock import Mock

from starlette.applications import Starlette
from starlette.datastructures import State
from starlette.requests import Request

from balladeer import Dialogue
from balladeer import Page
from balladeer import Presenter
from balladeer import Session
from balladeer import StoryBuilder


class ActionButtonTests(unittest.TestCase):

    @staticmethod
    def mock_request():
        request = Mock(spec=Request)
        request.app = Mock(spec=Starlette)
        request.app.state = Mock(spec=State)
        request.app.state.static = pathlib.Path(".")
        request.app.state.presenter = Presenter()
        return request

    class TestStory(StoryBuilder):
        pass

    def test_command_form(self):

        story = self.TestStory(
            Dialogue("<> Perhaps it's time to go to bed?"),
        )

        page = Page()
        request = self.mock_request()
        endpoint = Session(dict(type="http"), None, None)

        with story.turn() as turn:
            page = endpoint.compose(request, page, story, turn)
        lines = page.html.splitlines()

        command_form = next(
            (i for i in lines if i.startswith("<form") and 'name="ballad-command-form"' in i),
            None
        )
        self.assertTrue(command_form, lines)

    def test_code_implies_action(self):

        story = self.TestStory(
            Dialogue("<> Perhaps it's time to `go to bed`?"),
        )

        page = Page()
        request = self.mock_request()
        endpoint = Session(dict(type="http"), None, None)
        with story.turn() as turn:
            page = endpoint.compose(request, page, story, turn)
        lines = page.html.splitlines()

        command_form = next(
            (i for i in lines if i.startswith("<form") and 'name="ballad-command-form"' in i),
            None
        )
        self.assertTrue(command_form, lines)

        action_button = next(
            (i for i in lines if "<button" in i and 'form="ballad-action-form-go-to-bed"' in i),
            None
        )
        self.assertTrue(action_button, lines)
        action_form = next(
            (i for i in lines if i.startswith("<form") and 'id="ballad-action-form-go-to-bed"' in i),
            None
        )
        self.assertTrue(action_form, lines)



#!/usr/bin/env python
# encoding: utf8

# Copyright 2024 D E Haynes

# This file is part of balladeer.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import unittest

import balladeer
from balladeer import discover_assets
from balladeer import Entity
from balladeer import Loader
from balladeer import Resident
from balladeer import WorldBuilder
from busker.stager import Stager


class ResidentTests(unittest.TestCase):

    Spot = enum.Enum("Spot", {"kitchen": ["kitchen"], "hall": ["hallway"], "cloaks": ["cloakroom", "toilet"]})

    class TestResident(Resident):
        pass

    class TestWorld(WorldBuilder):

        def build(self, *args, **kwargs):
            yield Entity(type="Rukus").set_state(4)
            yield Entity(type="Focus").set_state(1, ResidentTests.Spot.kitchen)
            yield Entity(type="Focus").set_state(3)
            yield Entity(type="Focus").set_state(2)

    def test_is_resident(self):
        Colour = enum.Enum("Colour", ["red", "blue", "green", "yellow"])

        selector = {
            "states": [
                "colour.green",
                "spot.kitchen",
                "spot.cloaks",
            ]
        }

        drama = self.TestResident(selector=selector)
        self.assertIsInstance(drama.selector["states"], set)

        self.assertTrue(drama.is_resident(Colour.green))
        self.assertFalse(drama.is_resident(Colour.red))

        self.assertFalse(drama.is_resident(Colour.red, self.Spot.kitchen))
        self.assertTrue(drama.is_resident(Colour.green, self.Spot.kitchen))

    def test_scripts(self):
        selector = {
            "paths": [
                "balladeer/examples/ex_1[01]_*/*.scene.toml"
            ],
        }

        assets = discover_assets(balladeer, "examples")
        scenes = assets.get(Loader.Scene, [])

        drama = self.TestResident(selector=selector)
        scripts = drama.scripts(scenes)

        witness = set(i.path.parts[-3:] for i in scripts)
        self.assertEqual(
            witness,
            set((
                ("examples", "ex_10_animate_media", "argument.scene.toml"),
                ("examples", "ex_11_inventory_compass", "cloakroom.scene.toml"),
                ("examples", "ex_11_inventory_compass", "bar.scene.toml"),
                ("examples", "ex_11_inventory_compass", "foyer.scene.toml"),
            ))
        )

    def test_focus_no_selector(self):
        world = self.TestWorld()
        drama = self.TestResident(world=world)

        focus = drama.focus
        self.assertEqual(focus.get_state(), 3)
        self.assertIn("Focus", focus.types)

    def test_focus_with_selector(self):
        world = self.TestWorld()
        drama = self.TestResident(world=world, selector=dict(states=["spot.kitchen"]))

        focus = drama.focus
        self.assertEqual(focus.get_state(), 1)
        self.assertIn("Focus", focus.types)




#!/usr/bin/env python
#   -*- encoding: UTF-8 -*-
#
#   © 2008 Thuswise Ltd
#
#   This file is part of the Balladeer distribution.
#
#   Balladeer is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Balladeer is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Balladeer.  If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal
from math import sqrt, pi, cos, sin
import sys
import types

try:
    from types import FloatType
    from types import IntType
    from types import TupleType
except ImportError:
    FloatType = float
    IntType = int
    TupleType = tuple

import unittest

from balladeer.lite.homogeneous import cross
from balladeer.lite.homogeneous import dot
from balladeer.lite.homogeneous import Homogeneous
from balladeer.lite.homogeneous import maxpick
from balladeer.lite.homogeneous import minpick
from balladeer.lite.homogeneous import normalise
from balladeer.lite.homogeneous import point
from balladeer.lite.homogeneous import premultiply
from balladeer.lite.homogeneous import vector


class FactoryTester(unittest.TestCase):
    """
    Contains test cases for Homogeneous factory functions
    """

    def test_point_empty(self):
        obj = point()
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 1)
        self.assertEqual(obj, (1, ))
        self.assertEqual(type(obj[-1]), IntType)

    def test_point_int(self):
        obj = point(1, 2, 3)
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        self.assertEqual(obj, (1, 2, 3, 1))
        self.assertEqual(type(obj[-1]), IntType)

    def test_point_float(self):
        obj = point(1.1, 2.2, 3.3)
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        for a, b in zip(obj, (1.1, 2.2, 3.3, 1.0)):
            self.assertAlmostEqual(a, b)
        self.assertEqual(type(obj[-1]), FloatType)

    def test_point_decimal(self):
        obj = point(Decimal("1.1"), Decimal("2.2"), Decimal("3.3"))
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        self.assertEqual(obj[-1], Decimal("1.0"))
        self.assertEqual(type(obj[-1]), Decimal)

    def test_vector_empty(self):
        obj = vector()
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 1)
        self.assertEqual(obj, (0, ))
        self.assertEqual(type(obj[-1]), IntType)

    def test_vector_int(self):
        obj = vector(1, 2, 3)
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        self.assertEqual(obj, (1, 2, 3, 0))
        self.assertEqual(type(obj[-1]), IntType)

    def test_vector_float(self):
        obj = vector(1.1, 2.2, 3.3)
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        for a, b in zip(obj, (1.1, 2.2, 3.3, 0.0)):
            self.assertAlmostEqual(a, b)
        self.assertEqual(type(obj[-1]), FloatType)

    def test_vector_decimal(self):
        obj = vector(Decimal("1.1"), Decimal("2.2"), Decimal("3.3"))
        self.assertTrue(isinstance(obj, Homogeneous))
        self.assertTrue(isinstance(obj, TupleType))
        self.assertEqual(len(obj), 4)
        self.assertEqual(obj[-1], Decimal("0.0"))
        self.assertEqual(type(obj[-1]), Decimal)


class ArithmeticTester(unittest.TestCase):
    """
    Contains test cases for arithmetical operations on Homogeneous
    objects
    """

    def test_sub_vectors(self):
        x = vector(1.1, 2.2, 3.3)
        y = vector(0.5, 0.6, 0.7)
        z = x - y
        for a, b in zip(z, (0.6, 1.6, 2.6, 0.0)):
            self.assertAlmostEqual(a, b)

    def test_sub_points(self):
        x = point(1.1, 2.2, 3.3)
        y = point(0.5, 0.6, 0.7)
        z = x - y
        for a, b in zip(z, (0.6, 1.6, 2.6, 0.0)):
            self.assertAlmostEqual(a, b)

    def test_sub_mixed(self):
        x = point(1.1, 2.2, 3.3)
        y = vector(0.5, 0.6, 0.7)
        z = x - y
        for a, b in zip(z, (0.6, 1.6, 2.6, 1.0)):
            self.assertAlmostEqual(a, b)

    def test_add_vectors(self):
        x = vector(1.1, 2.2, 3.3)
        y = vector(0.5, 0.6, 0.7)
        z = x + y
        for a, b in zip(z, (1.6, 2.8, 4.0, 0.0)):
            self.assertAlmostEqual(a, b)

    def test_add_mixed(self):
        x = point(1.1, 2.2, 3.3)
        y = vector(0.5, 0.6, 0.7)
        z = x + y
        for a, b in zip(z, (1.6, 2.8, 4.0, 1.0)):
            self.assertAlmostEqual(a, b)

    def test_mult_vector_by_int(self):
        z = vector(1.1, 2.2, 3.3) * 3
        for a, b in zip(z, vector(3.3, 6.6, 9.9)):
            self.assertAlmostEqual(a, b)

    def test_mult_vector_by_float(self):
        z = 3.3 * vector(1, 2, 3)
        for a, b in zip(z, vector(3.3, 6.6, 9.9)):
            self.assertAlmostEqual(a, b)

    def test_div_vectors(self):
        z = vector(3.3, 6.6, 9.9) / 3
        for a, b in zip(z, vector(1.1, 2.2, 3.3)):
            self.assertAlmostEqual(a, b)

    def test_mult_points(self):
        p = point(1.1, 2.2, 3.3)
        self.assertRaises(
            NotImplementedError,
            p.__mul__, 3)

    def test_dot_int(self):
        """ Hill and Kelley 3rd Ed., example 4.3.1 """
        self.assertEqual(
            dot(vector(2, 3, 1), vector(0, 4, -1)),
            11
        )
        self.assertEqual(
            dot(vector(22, 2, 7), vector(12, -9, 11)),
            323
        )

    def test_dot_float(self):
        """ Hill and Kelley 3rd Ed., example 4.3.1 """
        self.assertEqual(
            dot(vector(2, 2, 2, 2), vector(4, 1, 2, 1.1)),
            16.2
        )
        self.assertEqual(
            dot(vector(169, 0, 43), vector(0, -375.3, 0)),
            0
        )

    def test_cross_int(self):
        """ Hill and Kelley 3rd Ed., example 4.4.3 """
        self.assertEqual(
            cross(vector(1, 3, -2), vector(0, 2, 2)),
            vector(10, -2, 2)
        )

    def test_gt_of_points(self):
        """ Test component-wise greater-than """
        self.assertTrue(point(1, 0, 0) > point(0, 0, 0))
        self.assertTrue(point(0, 1, 0) > point(0, 0, 0))
        self.assertTrue(point(0, 0, 1) > point(0, 0, 0))
        self.assertTrue(point(0, 0, 0) > point(-1, 0, 0))
        self.assertTrue(point(0, 0, 0) > point(0, -1, 0))
        self.assertTrue(point(0, 0, 0) > point(0, 0, -1))

    def test_lt_of_points(self):
        """ Test component-wise less-than """
        self.assertTrue(point(0, 0, 0) < point(1, 0, 0))
        self.assertTrue(point(0, 0, 0) < point(0, 1, 0))
        self.assertTrue(point(0, 0, 0) < point(0, 0, 1))
        self.assertTrue(point(-1, 0, 0) < point(0, 0, 0))
        self.assertTrue(point(0, -1, 0) < point(0, 0, 0))
        self.assertTrue(point(0, 0, -1) < point(0, 0, 0))

    def test_max_of_points(self):
        """ Test component-wise maximum """
        self.assertEqual(
            max(point(1, 0, 0), point(0, 0, 0)),
            point(1, 0, 0)
        )
        self.assertEqual(
            max(point(0, -1, 0), point(0, 0, 0)),
            point(0, 0, 0)
        )

    def test_min_of_points(self):
        """ Test component-wise minimum """
        self.assertEqual(
            min(point(0, 0, 0), point(1, 0, 0)),
            point(0, 0, 0)
        )
        self.assertEqual(
            min(point(0, -1, 0), point(0, 0, 0)),
            point(0, -1, 0)
        )

    def test_maxpick_of_points(self):
        """ Test component-wise maximum """
        self.assertEqual(
            maxpick(point(1, 0, 0), point(0, 1, 0)),
            point(1, 1, 0)
        )
        self.assertEqual(
            maxpick(point(0, -1, 1), point(0, 0, 0)),
            point(0, 0, 1)
        )

    def test_minpick_of_points(self):
        """ Test component-wise minimum """
        self.assertEqual(
            minpick(point(0, 0, -1), point(1, 0, 0)),
            point(0, 0, -1)
        )
        self.assertEqual(
            minpick(point(0, -1, 1), point(0, 0, 0)),
            point(0, -1, 0))

    def test_magnitude_of_vectors(self):
        """ Test calculation of vector magnitudes """
        self.assertAlmostEqual(vector(1, 1).magnitude, sqrt(2))
        self.assertAlmostEqual(vector(1, sqrt(3)).magnitude, 2)
        self.assertAlmostEqual(vector(3, 4).magnitude, 5)

    def test_magnitude_of_points(self):
        """ Test calculation of point vector magnitudes """
        self.assertAlmostEqual(point(1, 1).magnitude, sqrt(2))
        self.assertAlmostEqual(point(1, sqrt(3)).magnitude, 2)
        self.assertAlmostEqual(point(3, 4).magnitude, 5)

    def test_normalise_of_vector(self):
        """ Test normalisation of vectors """
        self.assertAlmostEqual(normalise(vector(1, 1)).magnitude, 1)
        self.assertAlmostEqual(
            normalise(vector(1, sqrt(3))).magnitude,
            1)
        self.assertAlmostEqual(normalise(vector(3, 4)).magnitude, 1)

    def test_premultiply_of_short_tuple(self):
        """ Check behaviour when premultiply is passed a tuple, rather
        than a Homogeneous object """
        # from Hill and Kelley 3rd Ed., ex 5.3.2
        c, s = 0.866, 0.5
        self.assertRaises(
            AssertionError,
            premultiply,
            (3, 1, 4), (c, 0, s, 0), (0, 1, 0, 0),
            (-s, 0, c, 0), (0, 0, 0, 1)
        )

    def test_premultiply_of_full_tuple(self):
        """ Check behaviour when premultiply is passed a tuple, rather
        than a Homogeneous object """
        # from Hill and Kelley 3rd Ed., ex 5.3.2
        c, s = 0.866, 0.5
        result = premultiply(
            (3, 1, 4, 1), (c, 0, s, 0), (0, 1, 0, 0),
            (-s, 0, c, 0), (0, 0, 0, 1))
        self.assertEqual(len(result), 4)
        self.assertAlmostEqual(result[0], 4.6, 2)
        self.assertAlmostEqual(result[1], 1.0, 2)
        self.assertAlmostEqual(result[2], 1.964, 2)
        self.assertAlmostEqual(result[3], 1.0, 2)
        self.assertEqual(type(result), Homogeneous)

    def test_premultiply_of_point(self):
        """ Check behaviour when premultiply is passed a point """
        # from Hill and Kelley 3rd Ed., ex 5.3.2
        c, s = 0.866, 0.5
        result = premultiply(
            point(3, 1, 4), (c, 0, s, 0), (0, 1, 0, 0),
            (-s, 0, c, 0), (0, 0, 0, 1)
        )
        self.assertEqual(len(result), 4)
        self.assertAlmostEqual(result[0], 4.6, 2)
        self.assertAlmostEqual(result[1], 1.0, 2)
        self.assertAlmostEqual(result[2], 1.964, 2)
        self.assertAlmostEqual(result[3], 1.0, 2)
        self.assertEqual(type(result), Homogeneous)

    def test_premultiply_of_vector(self):
        """ Check behaviour when premultiply is passed a vector """
        # from Hill and Kelley 3rd Ed., ex 5.3.2
        c, s = 0.866, 0.5
        result = premultiply(
            vector(3, 1, 4), (c, 0, s, 0), (0, 1, 0, 0),
            (-s, 0, c, 0), (0, 0, 0, 1)
        )
        self.assertEqual(len(result), 4)
        self.assertAlmostEqual(result[0], 4.6, 2)
        self.assertAlmostEqual(result[1], 1.0, 2)
        self.assertAlmostEqual(result[2], 1.964, 2)
        self.assertAlmostEqual(result[3], 0.0, 2)
        self.assertEqual(type(result), Homogeneous)



#!/usr/bin/env python
# encoding: utf8

# Copyright 2024 D E Haynes

# This file is part of balladeer.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import enum
import operator
import warnings
import sys

from balladeer.lite.compass import Compass
from balladeer.lite.drama import Drama
from balladeer.lite.entity import Entity
from balladeer.lite.types import Fruition
from balladeer.lite.types import Grouping
from balladeer.lite.loader import Loader
from balladeer.lite.compass import MapBuilder
from balladeer.lite.resident import Resident
from balladeer.lite.speech import Speech
from balladeer.lite.storybuilder import StoryBuilder
from balladeer.lite.compass import Transit
from balladeer.lite.compass import Traffic
from balladeer.lite.world import WorldBuilder
from busker.stager import Stager


class StoryStager(StoryBuilder):

    states = [Compass, Fruition, Traffic]
    types = [Drama, Entity, Transit]

    @classmethod
    def item_state(cls, spec: str | int, pool: list[enum.Enum] = [], default=0):
        try:
            name, value = spec.lower().split(".")
        except AttributeError:
            return spec

        lookup = {typ.__name__.lower(): typ for typ in pool + cls.states}

        try:
            cls = lookup[name]
        except KeyError:
            return default

        try:
            return cls[value]
        except KeyError:
            try:
                return cls[value.upper()]
            except KeyError:
                return default

    @classmethod
    def item_type(cls, name: str, default=Entity):
        name = name or ""
        return {
            typ.__name__.lower(): typ for typ in cls.types
        }.get(name.lower(), default)

    def __init__(
        self,
        *speech: tuple[Speech],
        config=None,
        assets: Grouping = Grouping(list),
        **kwargs,
    ):
        staging = [i.data for i in assets[Loader.Staging]]
        self.stager = Stager(staging).prepare()

        spots = self.stager.gather_state()
        m = MapBuilder(spots=spots)
        world = WorldBuilder(map=m, config=config, assets=assets)
        super().__init__(*speech, config=config, assets=assets, world=world, **kwargs)

    def __deepcopy__(self, memo):
        speech = copy.deepcopy(self.speech, memo)
        config = copy.deepcopy(self.config, memo)
        assets = copy.deepcopy(self.assets, memo)
        rv = self.__class__(*speech, config=config, assets=assets)
        return rv

    def make(self, **kwargs):
        self.drama: dict[tuple[str, str], Drama] = {
            (realm, name): self.build(realm, name) for (realm, name) in self.stager.puzzles
        } or {(None, None): Drama(*self.speech, config=self.config, world=self.world)}

    def build(self, realm: str, name: str, **kwargs):
        pool = [self.world.map.home, self.world.map.into, self.world.map.exit, self.world.map.spot]
        puzzle = self.stager.gather_puzzle(realm, name)
        drama_type = self.item_type(puzzle.get("type"), default=Resident)
        states = [self.item_state(f"{k}.{v}", pool=pool) for k, v in puzzle.get("init", {}).items()]
        drama = drama_type(
            *self.speech,
            config=self.config,
            world=self.world,
            name=puzzle.get("name"),
            type=drama_type.__name__,
            links=set(target for transition in puzzle.get("chain", {}).values() for target in transition),
            sketch=puzzle.get("sketch", ""),
            aspect=puzzle.get("aspect", ""),
            revert=puzzle.get("revert", ""),
            selector=puzzle.get("selector", {}),
        ).set_state(0, *states)

        for item in puzzle.get("items"):
            item_names = [i for i in [item.get("name")] + item.get("names", []) if i]
            item_types = list(filter(None, [item.get("type")] + item.get("types", [])))
            item_type = next((self.item_type(i) for i in item_types), Entity)
            states = [self.item_state(i, pool=pool) for i in item.get("states", [])]
            entity = item_type(
                names=item_names,
                types={i for i in item_types if i != item_type.__name__},
                links={puzzle.get("name")},
                sketch=item.get("sketch", ""),
                aspect=item.get("aspect", ""),
                revert=item.get("revert", ""),
            ).set_state(*states)

            if item_type is Transit:
                self.world.map.transits.append(entity)
            else:
                self.world.entities.append(entity)

        self.world.typewise = Grouping.typewise(self.world.entities)
        return drama

    @property
    def context(self):
        active = [self.drama[(realm, name)] for realm, name in self.stager.active]

        drama = sorted(
            [d for d in active if getattr(d, "focus", True)],
            key=operator.attrgetter("state"),
            reverse=True
        )
        drama = drama or active
        return next(iter(drama), None)

    def monitor_context(
        self,
        realm: str,
        name: str,
        drama: Drama,
        terminal={Fruition.withdrawn, Fruition.defaulted, Fruition.cancelled, Fruition.completion}
    ):
        pool = [self.world.map.home, self.world.map.into, self.world.map.exit, self.world.map.spot]
        if (state := drama.get_state(Fruition)) in terminal:
            for event in self.stager.terminate(realm, name, state.name):
                state = self.item_state(event.payload, pool=pool)
                if isinstance(event.target, str):
                    self.drama[event.realm, event.target].set_state(state)
                else:
                    entities = [
                        entity.set_state(state)
                        for entity in drama.ensemble
                        if set(event.target) <= entity.types
                    ]

    def turn(self, *args, **kwargs):
        for realm, name in self.stager.active.copy():
            drama = self.drama[(realm, name)]
            self.monitor_context(realm, name, drama)

        try:
            self.context.interlude(*args, **kwargs)
        except AttributeError:
            warnings.warn("Turn exhausted context")
        return self



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple
import importlib.resources
from importlib.resources import Package
import mimetypes
from pathlib import Path
import re
import tomllib
from typing import Mapping

from balladeer.lite.types import Grouping
from busker.stager import Stager


class Loader:
    Asset = namedtuple("Asset", ["resource", "path", "type", "stats"], defaults=[None, None])
    Scene = namedtuple(
        "Scene",
        ["text", "tables", "resource", "path", "stats"],
        defaults=[None, None, None],
    )
    Storage = namedtuple("Storage", ["resource", "path", "stats"], defaults=[None])
    Staging = namedtuple("Staging", ["text", "data", "resource", "path", "stats"], defaults=[None, None, None])
    Structure = namedtuple("Structure", ["text", "data", "resource", "path", "stats"], defaults=[None, None, None])

    @staticmethod
    def discover(
        package: [Package | Path],
        resource=".",
        suffixes={
            ".db": Storage,
            ".scene.toml": Scene,
            ".stage.toml": Staging,
            ".toml": Structure,
        },
        avoid=["tmp", "__pycache__", "node_modules"],
        ignore=[re.compile("^test_.*")],
    ):
        if isinstance(package, Path):
            paths = list(package.iterdir()) if package.is_dir() else [package]
        else:
            paths = list(importlib.resources.files(package).joinpath(resource).iterdir())

        for path in paths:
            if path.is_dir() and path.name in avoid:
                continue
            elif path.is_dir():
                yield from Loader.discover(path, resource, suffixes, avoid, ignore)
            elif any(i.match(path.name) for i in ignore):
                continue

            typ, _ = mimetypes.guess_type(path)
            if typ and typ != "text/x-python":
                with importlib.resources.as_file(path) as f:
                    yield Loader.Asset(resource, path, typ, f.stat())

            typ = suffixes.get("".join(path.suffixes))
            if typ in (Loader.Scene, Loader.Structure):
                with importlib.resources.as_file(path) as f:
                    text = f.read_text(encoding="utf8")
                    data = Loader.read_toml(text)
                    yield typ(text, data, resource, path, f.stat())
            elif typ == Loader.Staging:
                with importlib.resources.as_file(path) as f:
                    text = f.read_text(encoding="utf8")
                    data = next(Stager.load(text))
                    yield typ(text, data, resource, path, f.stat())
            elif typ:
                with importlib.resources.as_file(path) as f:
                    yield typ(resource, path, f.stat())

    @staticmethod
    def ignore_style(asset: Asset, *args) -> bool:
        if any(i in str(asset.path) for i in args):
            return True
        return "style" not in str(asset.path)

    @staticmethod
    def stage(assets: Grouping[str, list[Asset]], *args, predicate=None) -> Grouping[str, list[Asset]]:
        assert isinstance(assets, Mapping), type(assets)
        predicate = predicate or Loader.ignore_style

        rv = assets.copy()
        for key in assets:
            rv[key] = [i for i in assets[key] if predicate(i, *args)]

        return rv

    @staticmethod
    def read_toml(text: str, resource="", path=None):
        try:
            return tomllib.loads(text)
        except tomllib.TOMLDecodeError as e:
            return {}



#!/usr/bin/env python
# encoding: utf8

# Copyright 2024 D E Haynes

# This file is part of balladeer.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import enum
import unittest

import balladeer
from balladeer import Compass
from balladeer import discover_assets
from balladeer import Drama
from balladeer import Entity
from balladeer import Fruition
from balladeer import Grouping
from balladeer import Loader
from balladeer import Resident
from balladeer import StoryStager
from balladeer import Stager
from balladeer import Traffic
from balladeer import Transit

import busker
from busker.test.test_stager import StagerTests


class StoryTests(unittest.TestCase):

    class TestStory(StoryStager):
        types = [Drama, Entity, Resident, Transit]

    def setUp(self):
        assets = discover_assets(busker, "test", ignore=[])
        self.assertTrue(assets[Loader.Staging])
        self.story = self.TestStory(assets=assets)

    def test_make(self):
        self.assertIsInstance(getattr(self.story, "stager"), Stager)
        self.assertTrue(self.story.stager.realms)
        self.assertTrue(self.story.stager.active)

        self.assertTrue(issubclass(self.story.world.map.spot, enum.Enum))

        self.assertTrue(self.story.world.map.transits)
        self.assertFalse(
            any(len(i.names) > 1 for i in self.story.world.map.transits),
            self.story.world.map.transits
        )
        self.assertEqual(len([i for i in self.story.world.map.transits if i.get_state(Traffic)]), 9)
        self.assertEqual(len([i for i in self.story.world.map.transits if i.get_state(Compass)]), 8)
        self.assertGreater(len(self.story.world.map.spot), 1, list(self.story.world.map.spot))

        self.assertFalse(
            any(i.get_state(Traffic) for i in self.story.context.ensemble),
            self.story.context.ensemble
        )

    def test_story_copy_drama(self):
        b = copy.deepcopy(self.story)
        self.assertNotEqual(self.story.uid, b.uid, vars(self.story))
        self.assertIsNot(self.story.director, b.director, vars(self.story.director))

        self.assertTrue(self.story.drama)
        self.assertTrue(self.story.context)

        witness = []
        for drama in self.story.drama.values():
            self.assertEqual(sorted(set(drama.names)), sorted(drama.names), drama.names)
            with self.subTest(a=self.story, b=b, drama=drama):
                self.assertNotIn(
                    id(drama), [id(i) for i in b.drama.values()],
                    b.drama
                )
                witness.append(drama)

        self.assertTrue(witness)

    def test_story_copy_map(self):
        a = copy.deepcopy(self.story)
        b = copy.deepcopy(self.story)

        self.assertEqual(list(a.drama), list(b.drama))

        for entity in a.world.entities:
            with self.subTest(a=a, b=b, entity=entity):
                self.assertFalse(any(entity.names is i.names for i in b.world.entities), entity.names)
                self.assertFalse(any(entity.states is i.states for i in b.world.entities))
                self.assertFalse(any(entity.types is i.types for i in b.world.entities))

        self.assertTrue(a.world.map)
        self.assertTrue(b.world.map)

        self.assertIsNot(a.world, b.world)
        self.assertIsNot(a.world.map, b.world.map)

        self.assertGreater(len(a.world.map.spot), 1, list(a.world.map.spot))

        self.assertIsNot(a.world.map.spot, b.world.map.spot)
        self.assertEqual([str(i) for i in a.world.map.spot], [str(i) for i in b.world.map.spot])

        self.assertIsNot(a.world.map.exit, b.world.map.exit)
        self.assertEqual([str(i) for i in a.world.map.exit], [str(i) for i in b.world.map.exit])

        self.assertIsNot(a.world.map.into, b.world.map.into)
        self.assertEqual([str(i) for i in a.world.map.into], [str(i) for i in b.world.map.into])

        self.assertIsNot(a.world.map.home, b.world.map.home)
        self.assertEqual([str(i) for i in a.world.map.home], [str(i) for i in b.world.map.home])

        self.assertTrue(a.world.map.transits)
        self.assertTrue(b.world.map.transits, a.world.map.transits)
        for transit in a.world.map.transits:
            with self.subTest(a=a, b=b, transit=transit):
                self.assertFalse(any(transit.names is i.names for i in b.world.map.transits), transit.names)
                self.assertFalse(any(transit.states is i.states for i in b.world.map.transits))
                self.assertFalse(any(transit.types is i.types for i in b.world.map.transits))

    def test_story_copy_stager(self):
        s = copy.deepcopy(self.story)
        self.assertTrue(s.stager)
        self.assertEqual(len(self.story.stager.strands), len(s.stager.strands))
        for a in self.story.stager.strands.values():
            for b in s.stager.strands.values():
                with self.subTest(a=a, b=b):
                    self.assertIsNot(a, b)

    def test_story_turn(self):
        assets = Grouping.typewise([
            stage
            for rule in StagerTests.rules
            if (stage := Loader.Staging(text=rule, data=next(Stager.load(rule)))).data["realm"] == "busker"
        ])
        self.assertEqual(len(assets[Loader.Staging]), 2)
        s = self.TestStory(assets=assets)
        for n, _ in enumerate(set(s.stager.puzzles)):
            d = s.context
            with self.subTest(n=n, d=d):
                if n == 0:
                    self.assertEqual(d.name, "a")
                    self.assertTrue(hasattr(d, "selector"), d)
                    self.assertIsInstance(d.selector, dict)
                    self.assertEqual(set(d.selector), {"states", "paths"})
                    self.assertEqual(d.selector["states"], set(["spot.kitchen", "spot.hall"]))
                    d.set_state(Fruition.completion)
                    s.turn()
                    continue

                self.assertEqual(d.get_state(Fruition), Fruition.inception, d)
                d.set_state(Fruition.completion)
                if n + 1 < len(set(s.stager.puzzles)):
                    s.turn()
